-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("dailyTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LDT = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-DailyTooltip")
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2Daily", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon", text = ""})

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs = pairs
local format = string.format
local mod = mod
local GetTime = GetTime
local ToggleFrame = ToggleFrame
local CreateFrame = CreateFrame
local GetQuestResetTime = GetQuestResetTime
local GetMaxDailyQuests = GetMaxDailyQuests
local ldbInteractivityList = {
	["openOptions"] = LDT["Open tomQuest2 option frame"],
	["toggleQuestTracker"] = LDT["Toggle quests tracking frame"],
	["toggleAchievementTracker"] = LDT["Toggle achievements tracking frame"],
	["toggleAnnouncementModule"] = LDT["toggle chat announcement module"],
	["toggleQuestsLog"] = LDT["Toggle Blizzard quests log"],
	["toggleSyncPartyModule"] = LDT["Toggle party synchronization module"],
	["toggleQuestsTrackerModule"] = LDT["Toggle quests tracking module"],
	["toggleAchievementTrackerModule"] = LDT["Toggle achievements tracking module"],
	["nothing"] = LDT["Do nothing"],
}

-----------------
--- variables ---
-----------------
module.modName = LDT["Daily Quests Tooltip"]
local prdb, options
local dailytip
local counter = 0
local delay = 60
local maxDaily = GetMaxDailyQuests()
local dailytimer = CreateFrame("Frame")
local fd = 1/86400
local fh = 1/3600
local fm = 1/60

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedDailyQuests = true,
		showMaxDailyQuest = true,
		showRemainingTime = true,
		ldbMouseRightShift = "openOptions",
		ldbMouseRightControl = "openOptions",
		ldbMouseRightAlt = "openOptions",
		ldbMouseRight = "openOptions",
		ldbMouseLeftShift = "toggleQuestTracker",
		ldbMouseLeftControl = "toggleAnnouncementModule",
		ldbMouseLeftAlt = "toggleAchievementTracker",
		ldbMouseLeft = "toggleQuestsLog",
		ldbMouseMiddleShift = "toggleQuestTracker",
		ldbMouseMiddleControl = "toggleAnnouncementModule",
		ldbMouseMiddleAlt = "toggleAchievementTracker",
		ldbMouseMiddle = "toggleQuestsLog",
	}
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		module:updateLDBText()
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				counter = {
					type = "group",
					name = LDT["Daily quests counter"],
					args = {
						showNumCompletedDailyQuests = {
							name = LDT["Show number of completed daily quest"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						showMaxDailyQuest = {
							name = LDT["Show maximum daily quests number"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						showRemainingTime = {
							name = LDT["Show remaining time before end of day"],
							type = "toggle",
							width = "full",
							order = 3,
						},
					},
				},
				ldbInteraction = {
					type = "group",
					name = LDT["Daily quests counter interaction"],
					args = {
						ldbMouseRightShift = {
							name = LDT["Action on right shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 1,
						},
						ldbMouseRightControl = {
							name = LDT["Action on right control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 2,
						},
						ldbMouseRightAlt = {
							name = LDT["Action on right alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 3,
						},
						ldbMouseRight = {
							name = LDT["Action on right click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 4,
						},
						ldbMouseLeftShift = {
							name = LDT["Action on left shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 5,
						},
						ldbMouseLeftControl = {
							name = LDT["Action on left control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 6,
						},
						ldbMouseLeftAlt = {
							name = LDT["Action on left alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 7,
						},
						ldbMouseLeft = {
							name = LDT["Action on left click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 8,
						},
						ldbMouseMiddleShift = {
							name = LDT["Action on middle shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 9,
						},
						ldbMouseMiddleControl = {
							name = LDT["Action on middle control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 10,
						},
						ldbMouseMiddleAlt = {
							name = LDT["Action on middle alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 11,
						},
						ldbMouseMiddle = {
							name = LDT["Action on middle click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 12,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("dailyTooltip", defaults)
	prdb = self.db.profile
	self:SetEnabledState(core:getModuleEnabled("dailyTooltip"))
	core:registerModuleOptions("dailyTooltip", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "updateQuestsTracker", "updateLDBText")
	module:updateLDBText()
end

function dailytimer:OnUpdate(elapsed)
	counter = counter + elapsed
	if counter >= delay then
		counter = 0
		module:updateLDBText()
	end
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:updateLDBText(event)
	local remainingTime = GetQuestResetTime()
	if prdb.showNumCompletedDailyQuests then
		if prdb.showMaxDailyQuest then
			if prdb.showRemainingTime and remainingTime >= 0 and remainingTime <= 86400 then
				dataobj.text = format("|cff%s%s/%s|r |cff%s%d:%02d|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, maxDaily, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
			else
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, maxDaily)
			end
		elseif prdb.showRemainingTime and remainingTime >= 0 and remainingTime <= 86400 then
			dataobj.text = format("|cff%s%s|r |cff%s%d:%02d|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
		else
			dataobj.text = format("|cff%s%s|r", core:getColorFromCompletionHEX(core.numCompletedDailyQuests/maxDaily), core.numCompletedDailyQuests)
		end
	elseif prdb.showMaxDailyQuest then
		if prdb.showRemainingTime and remainingTime >= 0 and remainingTime <= 86400 then
			dataobj.text = format("|cffffffff%s|r |cff%s%d:%02d|r", maxDaily, core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
		else
			dataobj.text = maxDaily
		end	
	elseif prdb.showRemainingTime and remainingTime >= 0 and remainingTime <= 86400 then
		dataobj.text = format("|cff%s%d:%02d|r", core:getColorFromCompletionHEX(remainingTime*fd), remainingTime*fh, mod(remainingTime*fm, 60))
	else
		dataobj.text = ""
	end
end

local ldbOnClick = {
	openOptions = function()
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.Profiles)
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.tomQuest2)
	end,
	toggleQuestTracker = function()
		tQ2toggleQuestTracker()
	end,
	toggleAchievementTracker = function()
		tQ2toggleAchievementTracker()
	end,
	toggleAnnouncementModule = function()
		core:toggleAnnounce()
	end,
	toggleQuestsLog = function()
		ToggleFrame(QuestLogFrame)
	end,
	toggleSyncPartyModule = function()
		core:toggleSyncParty()
	end,
	toggleQuestsTrackerModule = function()
		core:toggleQuestsTracker()
	end,
	toggleAchievementTrackerModule = function()
		core:toggleAchievementTracker()
	end,
	nothing = function(value)
		return
	end,
}

function dataobj.OnClick(self, button)
	if button == "RightButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseRightShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseRightControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseRightAlt]()
		else
			ldbOnClick[prdb.ldbMouseRight]()
		end
	elseif button == "LeftButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftAlt]()
		else
			ldbOnClick[prdb.ldbMouseLeft]()
		end
	elseif button == "MiddleButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleAlt]()
		else
			ldbOnClick[prdb.ldbMouseMiddle]()
		end
	end
end
dailytimer:SetScript("OnUpdate", dailytimer.OnUpdate)