-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("tagQuestLog","AceEvent-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LTQL = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-TagQuestLog")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local tonumber, format, min = tonumber, string.format, math.min
local FauxScrollFrame_GetOffset, QuestLogListScrollFrame = FauxScrollFrame_GetOffset, QuestLogListScrollFrame
local GetNumQuestLogEntries, GetQuestLogTitle, GetQuestLink = GetNumQuestLogEntries, GetQuestLogTitle, GetQuestLink

-----------------
--- variables ---
-----------------
module.modName = LTQL["Quest Log Tag"]
local DUNGEON = LFG_TYPE_DUNGEON
local HEROIC = DUNGEON_DIFFICULTY2
local tags = {[ELITE] = "elite", [GROUP] = "group", [PVP] = "pvp", [RAID] = "raid", [DUNGEON] = "dungeon", [HEROIC] = "heroic", [DAILY] = "daily"}

function module:OnInitialize()
	self:SetEnabledState(core:getModuleEnabled("tagQuestLog"))
end

function module:OnEnable()
	self:SecureHook("QuestLog_Update")
end

function module:QuestLog_Update()
	local offset, numEntries, numQuests = FauxScrollFrame_GetOffset(QuestLogListScrollFrame), GetNumQuestLogEntries()
	for i=1, QUESTS_DISPLAYED do
		local index = i + offset
		if index > numEntries then break end
		local title, level, tag, group, header, _, complete, daily = GetQuestLogTitle(index)
		local questLogTitle, questTitleTag, questNormalText, questCheck = _G["QuestLogTitle"..i], _G["QuestLogTitle"..i.."Tag"], _G["QuestLogTitle"..i.."NormalText"], _G["QuestLogTitle"..i.."Check"]
		if questCheck then questCheck:Hide() end
		if not header then
			local uid = tonumber(GetQuestLink(index):match('quest:(%d+)'))
			tag, group, daily, level = tag or "", group or 0, daily or 0, level or 0
			if level > 0 or tag ~= "" or group > 0 or daily == 1 then
				title = format("[%s%s%s%s]%s", level > 0 and level or "", tag ~= "" and core.db.profile.quests[tags[tag]] or "", group > 0 and group or "", daily == 1 and (tag ~= DAILY and core.db.profile.quests[tags[DAILY]] or "") or "", title)
			end
			questLogTitle:SetText("  "..title)
			QuestLogDummyText:SetText("  "..title)
			if (not complete) or (complete == 0) then questTitleTag:SetText("") end
			questNormalText:SetWidth(min(QuestLogDummyText:GetWidth(), 260 - questTitleTag:GetWidth()))
			if uid and core.quests[uid] and core.quests[uid].tracked then
				questCheck:SetPoint("LEFT", questLogTitle, "LEFT")
				questCheck:Show()
			end
		end
	end
end